class ScaleLevel(object):
    """
    A value level used in scales.
    """

    def __init__(self, value, label, weight=0, **kwargs):
        self.value = value
        self.label = label
        self.weight = weight
        self.meta = kwargs

    def _as_tuple(self):
        return (self.weight, self.value, self.label)

    def __cmp__(self, other):
        return cmp(self._as_tuple(), other._as_tuple())

    def __hash__(self):
        return hash(self._as_tuple())

    def wrap(self, weight):
        return ScaleLevel(
            self.value, self.label, (weight, self.weight), **self.meta)


class BaseScale(object):
    """
    Base scale used to score raw values.
    """

    def get_levels(self):
        """
        Returns a list of all possible scale levels.
        """

        raise NotImplemented

    def get_level(self, value):
        """
        Retuns a ScaleLevel object or a number corresponding to the input value.
        """

        raise NotImplemented

    def score(self, value):
        """
        Returns the score for the given value.
        """

        level = self.get_level(value)
        if isinstance(level, ScaleLevel):
            return level.value
        return level


class StaticScale(BaseScale):
    """
    Scale that returns a static score regardless of the value.
    """

    def __init__(self, level):
        self.level = level

    def get_levels(self):
        """
        Returns a list of all possible scale levels.
        """

        return [self.level]

    def get_level(self, value):
        """
        Retuns a ScaleLevel object or a number corresponding to the input value.
        """

        return self.level


class BreaksScale(BaseScale):
    """
    A scale with predefined breaks.
    """

    def __init__(self, breaks, levels, right=True):
        if not breaks == sorted(breaks):
            raise ValueError('Breaks must be provided in increasing order')

        if len(breaks) + 1 != len(levels):
            raise IndexError('The number of levels must be one greater than '
                             'the number of breaks')

        self.breaks = breaks
        self.levels = levels
        self.right = right

    def get_levels(self):
        """
        Returns a list of all possible scale levels.
        """

        return self.levels

    def get_level(self, value):
        """
        Retuns a ScaleLevel object or a number corresponding to the input value.
        """

        breaks = self.breaks + [float('Inf')]
        for (break_value, level) in zip(breaks, self.levels):
            if (value < break_value) or (self.right and value == break_value):
                return level


class DictScale(BaseScale):
    """
    A scale based on a levels dictionary.
    """

    def __init__(self, levels, default=0):
        self.levels = levels
        self.default = default

    def get_levels(self):
        """
        Returns a list of all possible scale levels.
        """

        return self.levels.values()

    def get_level(self, value):
        """
        Retuns a ScaleLevel object or a number corresponding to the input value.
        """

        if value in self.levels:
            return self.levels[value]
        return self.default



import warnings
from numbers import Number
from cuuats.datamodel.domains import CodedValue, D
from cuuats.datamodel.field_values import DeferredValue
from cuuats.datamodel.scales import BaseScale, ScaleLevel
from cuuats.datamodel.query import RelatedManager


class BaseField(object):

    creation_index = 0
    default_storage = {}

    def __init__(self, label, **kwargs):
        self.label = label
        self.name = kwargs.get('name', None)
        self.db_name = kwargs.get('db_name', None)
        self.db_scale = kwargs.get('db_scale', None)
        self.db_precision = kwargs.get('db_precision', None)
        self.deferred = kwargs.get('deferred', False)
        self.required = kwargs.get('required', False)
        self.required_if = kwargs.get('required_if', None)
        self.domain_name = kwargs.get('domain_name', None)
        self.choices = list(kwargs.get('choices', []))
        self.storage = self.default_storage.copy()
        self.storage.update(kwargs.get('storage', {}))

        # Handle field ordering
        self.order = kwargs.get('order', 0)
        self.creation_index = BaseField.creation_index
        BaseField.creation_index += 1

    def __get__(self, instance, owner):
        # Retrieve deferred values if necessary.
        if isinstance(instance.values.get(self.name, None), DeferredValue):
            instance.get_deferred_values()

        # Get the current value from the instance.
        value = instance.values.get(self.name, None)

        # If this field has a coded values domain, set the description
        # for this value.
        if value is not None and self.domain_name:
            domain = instance.workspace.get_domain(self.domain_name)
            if domain.domainType == 'CodedValue':
                description = domain.codedValues.get(value, None)
                return CodedValue(value, description)

        return value

    def __set__(self, instance, value):
        # If this is a coded value, convert it back to a primative before
        # storing it.
        if isinstance(value, CodedValue):
            value = value.value

        # If the value representes a description for a coded values domain,
        # look up the value from the domain.
        elif isinstance(value, D):
            value = instance.workspace.get_coded_value(
                self.domain_name, value.description)

        instance.values[self.name] = value

    def __repr__(self):
        return '%s: %s' % (self.__class__.__name__, self.label)

    def register(self, workspace, feature_class, field_name, layer_name,
                 layer_fields):
        """
        Register this field with the workspace.
        """

        layer_field = layer_fields.get(self.db_name or field_name, None)
        if layer_field is None:
            warnings.warn('%s is not a field of %s' % (
                field_name, layer_name))
            return None

        self.name = field_name
        if self.db_name is None:
            self.db_name = field_name
        if layer_field.domain:
            self.domain_name = layer_field.domain
            domain = workspace.get_domain(layer_field.domain)
            if domain.domainType == 'CodedValue':
                self.choices.extend(domain.codedValues)
            elif hasattr(self, 'min') and hasattr(self, 'max'):
                self.min, self.max = domain.range
        if layer_field.scale:
            self.db_scale = layer_field.scale
        if layer_field.precision:
            self.db_precision = layer_field.precision
        return layer_field

    def validate(self, value):
        """
        Performs validation on the given value, and returns any error messages.
        """

        # Check for missing values in required fields.
        if self.required and value is None:
            return ['%s is missing' % (self.label,)]

        # Check that that value is in the choices, if specified.
        elif len(self.choices) > 0 and value is not None and \
                value not in self.choices:
            return ['%s is invalid' % (self.label,)]

        return []

    def round(self, value):
        """
        Rounds the given value to the number of decimal places it will have
        in the database.
        """

        if self.db_scale is not None and isinstance(value, Number):
            return round(value, self.db_scale)

        return value

    def has_changed(self, old, new):
        """
        Returns true if the new value is different from the old value.
        """

        # When the scale is set, we round to that number of decimal places
        # before comparing in order to determine whether the values will be
        # equal once saved in the database.
        return self.round(old) != self.round(new)

    def summarize(self, instance):
        """
        Returns the summary level for the given instance.
        """

        value = self.__get__(instance, None)
        return SummaryLevel(0, value, str(value))


class OIDField(BaseField):
    """
    Object ID field type.
    """

    def validate(self, value):
        """
        Performs validation on the given value, and returns any error messages.
        """

        # OIDs are always valid.
        return []


class GlobalIDField(BaseField):
    """
    Global ID field type.
    """

    def validate(self, value):
        """
        Performs validation on the given value, and returns any error messages.
        """

        # GlobalIDs are always valid.
        return []


class GeometryField(BaseField):
    """
    Geometry field type.
    """

    def __init__(self, label, **kwargs):
        super(GeometryField, self).__init__(label, **kwargs)
        self.db_name = kwargs.get('db_name', 'SHAPE@')
        self.deferred = kwargs.get('deferred', True)

    def register(self, workspace, feature_class, field_name, layer_name,
                 layer_fields):
        """
        Register this field with the workspace.
        """

        # TODO: Figure out how to check whether the geometry field exists,
        # since the db_name is a special ArcGIS keyword that is not returned
        # by ListFields.

        self.name = field_name
        if self.db_name is None:
            self.db_name = field_name

    def validate(self, value):
        """
        Performs validation on the given value, and returns any error messages.
        """

        # TODO: Add validation for geometries.
        return super(GeometryField, self).validate(value)


class StringField(BaseField):

    default_storage = {
        'field_type': 'TEXT',
        'field_length': 100
    }

    def validate(self, value):
        """
        Performs validation on the given value, and returns any error messages.
        """

        # Convert empty strings to None before performing validation.
        if isinstance(value, str) and value.strip() == '':
            value = None
        return super(StringField, self).validate(value)


class NumericField(BaseField):

    default_storage = {
        'field_type': 'FLOAT',
    }

    def __init__(self, name, **kwargs):
        super(NumericField, self).__init__(name, **kwargs)
        self.min = kwargs.get('min', None)
        self.max = kwargs.get('max', None)

    def validate(self, value):
        """
        Performs validation on the given value, and returns any error messages.
        """

        messages = super(NumericField, self).validate(value)
        if value is None:
            return messages

        if (self.min is not None and value < self.min) or \
                (self.max is not None and value > self.max):
            messages.append('%s out of range' % (self.label,))

        return messages


class BlobField(BaseField):
    """
    A field for storing blob data.
    """

    def __init__(self, label, **kwargs):
        super(BlobField, self).__init__(label, **kwargs)
        self.deferred = kwargs.get('deferred', True)


class CalculatedField(BaseField):

    def __init__(self, name, **kwargs):
        super(CalculatedField, self).__init__(name, **kwargs)

        # Overridden by subclasses
        self.condition = kwargs.get('condition', None)
        self.default = kwargs.get('default', None)
        self.scale = kwargs.get('scale', None)

    def _as_scale_level(self, level):
        if isinstance(level, ScaleLevel):
            return level
        return ScaleLevel(level, str(level))

    def __get__(self, instance, owner):
        if not instance.check_condition(self.condition):
            return self.default
        return self.calculate(instance)

    def __set__(self, instance, value):
        raise ValueError('Calculated fields cannot be set')

    def calculate(self, instance):
        """
        Calculate the value for this field based on the state of the instance.
        """

        # Overridden by subclasses
        return self.default

    def summarize(self, instance):
        """
        Returns the scale level for the given instance.
        """

        if not self.scale:
            raise AttributeError(
                'Weights fields must have a scale to be summarized')

        value = self.round(self.calculate(instance))
        return self._as_scale_level(self.scale.get_level(value))

    def get_levels(self):
        """
        Returns a sorted list of all possible scale levels.
        """

        if not self.scale:
            raise AttributeError(
                'Weights fields must have a scale to be summarized')

        return sorted(list(set(
            [self._as_scale_level(l) for l in self.scale.get_levels()])))


class MethodField(CalculatedField):

    default_storage = {
        'field_type': 'DOUBLE',
    }

    def __init__(self, name, **kwargs):
        super(MethodField, self).__init__(name, **kwargs)
        self.method_name = kwargs.get('method_name')

    def calculate(self, instance):
        """
        Calculate the value for this field based on the state of the instance.
        """

        return getattr(instance, self.method_name)(self.name)


class WeightsField(CalculatedField):

    default_storage = {
        'field_type': 'DOUBLE',
    }

    def __init__(self, name, **kwargs):
        super(WeightsField, self).__init__(name, **kwargs)
        self.weights = kwargs.get('weights')

    def _get_value(self, instance, field_name):
        return getattr(instance, field_name)

    def calculate(self, instance):
        """
        Calculate the value for this field based on the state of the instance.
        """

        if None in [self._get_value(instance, v) for v in self.weights.keys()]:
            return self.default
        else:
            return sum([self._get_value(instance, v)*w
                        for (v, w) in self.weights.items()])


class ScaleField(CalculatedField):

    default_storage = {
        'field_type': 'SHORT',
    }

    def __init__(self, name, **kwargs):
        super(ScaleField, self).__init__(name, **kwargs)
        self.value_field = kwargs.get('value_field')
        self.use_description = kwargs.get('use_description', False)

        if not isinstance(self.scale, (BaseScale, list, tuple)):
            raise TypeError('Scale must be a subclass of BaseScale or a '
                            'list containing tuples of conditions and scales')

    def _unpack_scales(self):
        if isinstance(self.scale, BaseScale):
            yield (None, self.scale, 0)
        else:
            for (index, scale_info) in enumerate(self.scale):
                if len(scale_info) == 3:
                    yield scale_info
                else:
                    condition, scale = scale_info
                    yield (condition, scale, index)

    def _get_scale_for(self, instance):
        for (condition, scale, weight) in self._unpack_scales():
            if instance.check_condition(condition):
                return (scale, weight)

        return (None, 0)

    def _get_value_for(self, instance):
        value = instance.eval(self.value_field)
        if self.use_description and isinstance(value, CodedValue):
            value = value.description
        return value

    def _as_scale_level(self, scale_weight, level):
        if isinstance(level, ScaleLevel):
            return level.wrap(scale_weight)
        return ScaleLevel(level, str(level), (scale_weight, 0))

    def calculate(self, instance):
        """
        Calculate the value for this field based on the state of the instance.
        """

        value = self._get_value_for(instance)
        scale, weight = self._get_scale_for(instance)
        if not scale:
            return self.default

        return scale.score(value)

    def summarize(self, instance):
        """
        Returns the summary level for the given instance.
        """

        value = self._get_value_for(instance)
        scale, scale_weight = self._get_scale_for(instance)
        if not scale:
            return ScaleLevel(self.default, str(self.default), (0, 0))

        return self._as_scale_level(scale_weight, scale.get_level(value))

    def get_levels(self):
        """
        Returns a sorted list of all possible scale levels.
        """

        levels = []
        for (condition, scale, weight) in self._unpack_scales():
            levels.extend(
                [self._as_scale_level(weight, l) for l in scale.get_levels()])
        return sorted(list(set(levels)))


class ForeignKey(BaseField):

    default_storage = {
        'field_type': 'SHORT',
    }

    def __init__(self, name, **kwargs):
        super(ForeignKey, self).__init__(name, **kwargs)
        # TODO: Allow origin_class to be the name of the class instead
        # of the actual class.
        self.origin_class = kwargs.get('origin_class', None)
        self.primary_key = kwargs.get('primary_key', None)
        self.related_name = kwargs.get('related_name', None)

    def register(self, workspace, feature_class, field_name, layer_name,
                 layer_fields):
        """
        Register this field with the workspace.
        """

        super(ForeignKey, self).register(
            workspace, feature_class, field_name, layer_name, layer_fields)

        if self.primary_key is None:
            self.primary_key = self.origin_class.fields.oid_field.name

        if self.related_name is None:
            self.related_name = feature_class.__name__.lower() + '_set'

        setattr(self.origin_class, self.related_name,
                RelatedManager(self.related_name, feature_class,
                               self.name, self.primary_key))

        self._set_related(feature_class)

    def _set_related(self, destination_class):
        """
        Set references to related feature classes.
        """

        for feature_class in (self.origin_class, destination_class):
            if feature_class.related_classes is None:
                feature_class.related_classes = {}

        self.origin_class.related_classes[destination_class.__name__] = \
            destination_class
        destination_class.related_classes[self.origin_class.__name__] = \
            self.origin_class

    def __get__(self, instance, owner):
        value = super(ForeignKey, self).__get__(instance, owner)
        if value is None:
            return None

        # Return the prefetched feature if there is one.
        prefetched_feature = instance._prefetch_cache.get(self.name, None)
        if prefetched_feature:
            return prefetched_feature

        # Otherwise, get the related feature from the database.
        return self.origin_class.objects.get({
            self.primary_key: value
        })

    def __set__(self, instance, value):
        # Clear prefetched feature for this relationship.
        if self.name in instance._prefetch_cache:
            del instance._prefetch_cache[self.name]

        # Allow setting using the primary key or the feature itself.
        if isinstance(value, self.origin_class):
            super(ForeignKey, self).__set__(getattr(value, self.primary_key))
        else:
            super(ForeignKey, self).__set__(value)



"""
Classes for working with domains.
"""


class Description(object):
    """
    A description associated with a coded value used in a coded values domain.
    """

    def __init__(self, description):
        if not isinstance(description, basestring):
            raise TypeError('Description must be a string')

        self.description = description

    def __eq__(self, other):
        if isinstance(other, Description):
            return self.description == other.description
        elif isinstance(other, CodedValue):
            return self.description == other.description
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

D = Description


class CodedValue(object):
    """
    A coded value with a description.
    """

    def __new__(cls, value, description):
        if isinstance(value, int):
            obj = int.__new__(IntCodedValue, value)
        elif isinstance(value, long):
            obj = long.__new__(LongCodedValue, value)
        elif isinstance(value, float):
            obj = float.__new__(FloatCodedValue, value)
        elif isinstance(value, str):
            obj = str.__new__(StringCodedValue, value)
        elif isinstance(value, unicode):
            obj = unicode.__new__(UnicodeCodedValue, value)
        else:
            raise TypeError(
                'Value must be of type int, long, float, str, or unicode')
        obj.description = description
        return obj

    @property
    def value(self):
        """
        Returns the value as a primative.
        """

        return type.mro(self.__class__)[2](self)


class IntCodedValue(CodedValue, int):
    """
    An integer coded value.
    """


class LongCodedValue(CodedValue, long):
    """
    A long integer coded value.
    """


class FloatCodedValue(CodedValue, float):
    """
    A floating point coded value.
    """


class StringCodedValue(CodedValue, str):
    """
    A string coded value.
    """


class UnicodeCodedValue(CodedValue, unicode):
    """
    A unicode coded value.
    """



import itertools
from collections import defaultdict
from cuuats.datamodel.exceptions import ObjectDoesNotExist, \
    MultipleObjectsReturned
from cuuats.datamodel.field_values import DeferredValue
from cuuats.datamodel.domains import D


class SQLCondition(object):

    OPERATORS = {
        'contains': 'LIKE',
        'eq': '=',
        'exact': 'IS',
        'gt': '>',
        'in': 'IN',
        'lt': '<',
        'gte': '>=',
        'lte': '<=',
    }

    def __init__(self, field_name, value, op_name=None):
        self.field_name = field_name
        self.value = value

        # Extract or intuit the operator.
        if op_name is None:
            if value is None:
                op_name = 'exact'
            else:
                op_name = 'eq'
        self.operator = self.OPERATORS[op_name]

    def __repr__(self):
        return '<SQLCondition: %s>' % (
            ' '.join([self.field_name, self.operator, str(self.value)]),)


class Q(object):

    def __init__(self, filters={}, **kwargs):
        self.operator = 'AND'
        self.negated = False
        self.rel_name = None
        self.children = []
        self._parse_children(filters.items() + kwargs.items())

    def __and__(self, other):
        return self._combine(other, 'AND')

    def __or__(self, other):
        return self._combine(other, 'OR')

    def __invert__(self):
        q = self._clone()
        q.negated = not self.negated
        return q

    def __len__(self):
        return len(self.children)

    def __repr__(self):
        child_str = ', '.join([str(c) for c in self.children])
        result = '%s[%s]: %s' % (self.rel_name or '', self.operator, child_str)
        if self.negated:
            result = 'NOT(%s)' % (result,)
        return '<Q:%s>' % (result,)

    def _parse_children(self, filters):
        for (field_info, value) in filters:
            rels, field_name, op_name = self._parse_field_info(field_info)
            parent = self
            for rel in rels:
                child = self._create()
                child.rel_name = rel
                parent.children.append(child)
                parent = child
            parent.children.append(SQLCondition(field_name, value, op_name))

    def _parse_field_info(self, field_info):
        op_name = None
        parts = field_info.split('__')
        if parts[-1] in SQLCondition.OPERATORS:
            op_name = parts.pop()
        field_name = parts.pop()
        return (parts, field_name, op_name)

    def _create(self, *args, **kwargs):
        return self.__class__(*args, **kwargs)

    def _clone(self, children=True):
        q = self._create()
        q.operator = self.operator
        q.negated = self.negated
        q.rel_name = self.rel_name
        if children:
            q.children = self.children[:]
        return q

    def _combine(self, other, op):
        q = self._create()
        q.children = [self._clone(), other._clone()]
        q.operator = op
        return q

    def _op_match(self, op):
        return op == self.operator or len(self) == 1

    def _can_merge(self, other, op):
        return self._op_match(op) and other._op_match(op) and \
            self.negated == other.negated and self.rel_name == other.rel_name

    def _can_append(self, other, op):
        return self._op_match(op) and not self.negated and \
            (self.rel_name is None or self.rel_name == other.rel_name)

    def _merge(self, other, op):
        q = self._clone()
        q.children.extend(other.children)
        q.operator = op
        return q

    def _append(self, other, op):
        q = self._clone()
        child = other._clone()
        if q.rel_name is not None:
            child.rel_name = None
        q.children.append(child)
        q.operator = op
        return q

    def _merge_or_append(self, other, op):
        if self._can_merge(other, op):
            return self._merge(other, op)
        elif self._can_append(other, op):
            return self._append(other, op)
        elif other._can_append(self, op):
            return other._append(self, op)
        return None

    def simplify(self):
        q = self._clone()
        q.children = [c.simplify() if isinstance(c, Q) else c
                      for c in q.children]

        # Merge/append children with/to each other.
        num_children = 0
        while num_children < len(q.children):
            num_children = len(q.children)
            for (a, b) in itertools.combinations(q.children, 2):
                if not isinstance(a, Q) or not isinstance(b, Q):
                    continue

                new_child = a._merge_or_append(b, q.operator)
                if new_child:
                    q.children = [new_child] + \
                        [c for c in q.children if c not in (a, b)]
                    break

        # Simplify the partent.
        parent_dirty = True
        while parent_dirty:
            parent_dirty = False
            # Merge children with the parent.
            for child in q.children[:]:
                if isinstance(child, Q) and q._can_merge(child, q.operator):
                    q.children = [c for c in q.children if c is not child]
                    q = q._merge(child, q.operator)
                    parent_dirty = True

            # Remove unnecessary parents.
            if len(q.children) == 1 and not q.negated and q.rel_name is None \
                    and isinstance(q.children[0], Q):
                q = q.children[0]
                parent_dirty = True

        return q


class SQLCompiler(object):

    def __init__(self, feature_class=None):
        self.feature_class = feature_class

    def compile(self, q, inner=False):
        q = q.simplify()
        where_parts = []
        feature_class, other_key, self_key = self._resolve_rel(q.rel_name)

        for child in q.children:
            if isinstance(child, Q):
                compiler = self.__class__(feature_class)
                where_parts.append(compiler.compile(child))
            else:
                where_parts.append(
                    self._compile_sql_condition(child, feature_class))

        sep = ' %s ' % (q.operator)
        where = sep.join(where_parts)

        if len(where_parts) > 1 and (q.negated or inner):
            where = '(%s)' % (where,)
        if q.negated:
            where = 'NOT %s' % (where,)
        if self_key is not None:
            where = '%s IN (SELECT %s FROM %s WHERE %s)' % (
                other_key, self_key, feature_class.name, where)

        return where

    def _compile_sql_condition(self, cond, feature_class):
        return ' '.join([
            self._resolve_field_name(cond.field_name, feature_class),
            cond.operator,
            self._to_string(cond.field_name, cond.value, cond.operator)])

    def _to_string(self, field_name, value, operator):
        if value is None:
            return 'NULL'

        elif isinstance(value, D):
            return self._to_string(
                field_name,
                self._resolve_coded_value(field_name, value), operator)

        elif isinstance(value, basestring):
            if operator == 'LIKE':
                return "'%%%s%%'" % (value,)
            return "'%s'" % (value,)

        elif isinstance(value, (list, tuple)):
            return '(%s)' % (', '.join(
                [self._to_string(field_name, v, operator) for v in value]), )

        # TODO: Deal with dates and other common types.
        return str(value)

    def _resolve_coded_value(self, field_name, d):
        field = self.feature_class.fields.get(field_name)
        return self.feature_class.workspace.get_coded_value(
            field.domain_name, d.description)

    def _resolve_field_name(self, field_name, feature_class):
        return feature_class.fields.get_db_name(field_name)

    def _resolve_rel(self, rel_name):
        if rel_name is None:
            return [self.feature_class, None, None]

        relation = self.feature_class.__dict__.get(rel_name)
        if isinstance(relation, RelatedManager):
            return [relation.destination_class,
                    self.feature_class.fields.get_db_name(
                        relation.primary_key),
                    relation.destination_class.fields.get_db_name(
                        relation.foreign_key)]

        # Field is a ForeignKey.
        return [relation.origin_class,
                relation.db_name,
                relation.origin_class.fields.get_db_name(relation.primary_key)]


class Query(object):

    def __init__(self, fields, compiler):
        self.fields = fields
        self.compiler = compiler
        self._where = None
        self._order_by = None
        self._group_by = None

    def clone(self):
        clone = self.__class__(self.fields, self.compiler)
        clone._where = self._where
        clone._order_by = self._order_by
        clone._group_by = self._group_by
        return clone

    def add_q(self, q):
        if self._where is None:
            self._where = q
        else:
            self._where = self._where & q

    def set_order(self, fields):
        self._order_by = []
        for field in fields:
            if isinstance(field, basestring):
                self._order_by.append([field, 'ASC'])
            else:
                self._order_by.append(field)

    def reverse_order(self):
        new_order = []
        for (field, direction) in self._order_by:
            if direction == 'ASC':
                new_order.append([field, 'DESC'])
            else:
                new_order.append([field, 'ASC'])
        self._order_by = new_order

    @property
    def where(self):
        if self._where is None:
            return None
        return self.compiler.compile(self._where)

    @property
    def prefix(self):
            return None

    @property
    def postfix(self):
        if not self._order_by:
            return None

        return 'ORDER BY %s' % (
            ', '.join(['%s %s' % (c, o) for (c, o) in self._order_by]))


class QuerySet(object):

    def __init__(self, feature_class, query=None):
        self.feature_class = feature_class
        self.query = query or self._make_query(feature_class)
        self._field_name_cache = None
        self._db_name_cache = None
        self._cache = None
        self._prefetch_rel = []
        self._prefetch_deferred = []

    def __len__(self):
        return self.count()

    def __iter__(self):
        self._fetch_all()
        return iter(self._cache)

    def _make_query(self, feature_class):
        fields = [f.db_name for f in feature_class.fields.values()
                  if not f.deferred]
        compiler = SQLCompiler(feature_class)
        query = Query(fields, compiler)
        oid_field = feature_class.fields.oid_field
        query.set_order([(oid_field.db_name, 'ASC')])
        return query

    def _make_q(self, *args, **kwargs):
        if args and isinstance(args[0], Q):
            return args[0] & Q(*args[1:], **kwargs)
        return Q(*args, **kwargs)

    def _fetch_all(self):
        if self._cache is None:
            self._cache = list(self.iterator())
            if self._cache:
                self._prefetch()

    def _prefetch(self):
        # Prefetch related features.
        for rel_name in self._prefetch_rel:
            rel = self.feature_class.__dict__.get(rel_name, None)
            if isinstance(rel, RelatedManager):
                self._prefetch_related_manager(rel_name, rel)
            elif rel.__class__.__name__ == 'ForeignKey':
                self._prefetch_foreign_key(rel_name, rel)
            else:
                raise AttributeError(
                    'Relationship %s does not exist.' % (rel_name,))

    def _prefetch_related_manager(self, rel_name, rel):
        # rel is a RelatedManager.
        destination = rel.destination_class

        pk_filter = '%s__in' % (rel.foreign_key,)
        pks = [getattr(f, rel.primary_key) for f in self._cache]
        dest_features = destination.objects.filter({pk_filter: pks})

        dest_map = defaultdict(list)
        for feature in dest_features:
            dest_map[feature.values.get(rel.foreign_key)].append(feature)

        for feature in self._cache:
            feature._prefetch_cache[rel_name] = dest_map[
                getattr(feature, rel.primary_key)]

    def _prefetch_foreign_key(self, rel_name, rel):
        # rel is a ForeignKey.
        origin = rel.origin_class

        fk_filter = '%s__in' % (rel.primary_key,)
        fks = [f.values.get(rel_name, None) for f in self._cache]
        fks = [fk for fk in fks if fk is not None]
        origin_features = origin.objects.filter({fk_filter: fks})

        origin_map = dict(
            [(getattr(f, rel.primary_key), f) for f in origin_features])

        for feature in self._cache:
            feature._prefetch_cache[rel_name] = origin_map.get(
                feature.values.get(rel_name), None)

    def _clone(self, preserve_cache=False):
        clone = self.__class__(self.feature_class, self.query.clone())
        clone._field_name_cache = self._field_name_cache
        clone._db_name_cache = self._db_name_cache
        clone._prefetch_rel = self._prefetch_rel

        if preserve_cache:
            clone._cache = self._cache

        return clone

    @property
    def _field_names(self):
        if self._field_name_cache is None:
            self._field_name_cache = self.feature_class.fields.keys()
        return self._field_name_cache

    @property
    def _db_names(self):
        if self._db_name_cache is None:
            self._db_name_cache = [
                f.db_name for f in self.feature_class.fields.values()]
        return self._db_name_cache

    def _feature(self, row):
        fields = zip(self._field_names, self._db_names)
        row_map = dict(zip(self.query.fields, row))
        values = [row_map.get(d, DeferredValue(f, d)) for (f, d) in fields]
        return self.feature_class(**dict(zip(self._field_names, values)))

    # Methods that return QuerySets
    def all(self):
        return self._clone(preserve_cache=True)

    def filter(self, *args, **kwargs):
        clone = self._clone()
        clone.query.add_q(self._make_q(*args, **kwargs))
        return clone

    def exclude(self, *args, **kwargs):
        clone = self._clone()
        clone.query.add_q(~self._make_q(*args, **kwargs))
        return clone

    def order_by(self, fields):
        clone = self._clone()
        clone.query.set_order(fields)
        return clone

    # Methods that do not return QuerySets
    def get(self, *args, **kwargs):
        clone = self._clone()
        clone.query.add_q(self._make_q(*args, **kwargs))
        clone._fetch_all()
        clone_length = len(clone)
        if clone_length == 1:
            return clone._cache[0]
        elif clone_length == 0:
            raise ObjectDoesNotExist(clone.query.where)
        raise MultipleObjectsReturned(clone.query.where)

    def get_or_create(self, *args, **kwargs):
        try:
            return self.get(*args, **kwargs)
        except ObjectDoesNotExist:
            return self.feature_class(*args, **kwargs)

    def count(self):
        # TODO: Investigate whether using selection would be faster in cases
        # where the results are not already cached.
        if self._cache is not None:
            return len(self._cache)

        return self.feature_class.workspace.count_rows(
            self.feature_class.name,
            self.query.where)

    def iterator(self, limit=None):
        for (row, cursor) in self.feature_class.workspace.iter_rows(
                self.feature_class.name, self.query.fields, False,
                self.query.where, limit, self.query.prefix,
                self.query.postfix):
            yield self._feature(row)

    def first(self):
        if self._cache is not None:
            if self._cache:
                return self._cache[0]
            return None

        results = list(self.iterator(limit=1))
        if results:
            return results[0]
        return None

    def last(self):
        if self._cache is not None:
            if self._cache:
                return self._cache[-1]
            return None

        clone = self._clone()
        clone.query.reverse_order()
        return clone.first()

    def summarize(self, summary_field_name, **kwargs):
        summary_field = self.feature_class.fields.get(summary_field_name, None)
        if not summary_field:
            raise KeyError('Invalid summary field name')

        levels = summary_field.get_levels()
        results = {}

        for level in levels:
            level_key = hash(level)
            results[level_key] = {}
            results[level_key].update(level.meta)
            results[level_key].update(dict(zip(kwargs.keys(), [0]*len(kwargs))))
            results[level_key]['count'] = 0
            results[level_key]['value'] = level.value
            results[level_key]['label'] = level.label

        for feature in self:
            level = summary_field.summarize(feature)
            level_key = hash(level)

            results[level_key]['count'] += 1
            for value_key, value_expr in kwargs.items():
                results[level_key][value_key] += feature.eval(value_expr)

        return [results[hash(l)] for l in levels]

    def aggregate(self, fields):
        return self.feature_class.workspace.summarize(
            self.feature_class.name,
            fields,
            self.query.where)

    def exists(self):
        if self._cache:
            return True
        return self.first() is not None

    def update(self):
        raise NotImplementedError('QuerySet updates are not yet supported')

    def delete(self):
        raise NotImplementedError('QuerySet deletions are not yet supported')

    def prefetch_related(self, *rels):
        for rel in rels:
            if rel not in self._prefetch_rel:
                self._prefetch_rel.append(rel)
        return self

    def prefetch_deferred(self, *field_names):
        for field_name in field_names:
            field = self.feature_class.fields.get(field_name, None)
            if field is None:
                raise AttributeError('%s does not have field "%s"' % (
                    self.feature_class.__name__, field_name))
            if field.db_name not in self.query.fields:
                self.query.fields.append(field.db_name)
        return self


class Manager(object):

    def __init__(self, queryset_class=QuerySet):
        self.queryset_class = queryset_class

    def __get__(self, instance, owner):
        if instance is not None:
            raise AttributeError('Manager is not accessible '
                                 'from feature instances')

        if owner.workspace is None:
            raise AttributeError(
                'Feature class %s must be registered '
                'before it can be queried' % (owner.__name__,))

        return self.queryset_class(owner)


class RelatedManager(Manager):

    def __init__(self, name, destination_class, foreign_key, primary_key,
                 queryset_class=QuerySet):
        super(RelatedManager, self).__init__(queryset_class)
        self.name = name
        self.destination_class = destination_class
        self.foreign_key = foreign_key
        self.primary_key = primary_key

    def __get__(self, instance, owner):
        if instance is None:
            raise AttributeError('Related Manager is only accessible '
                                 'from feature instances')

        if self.destination_class.workspace is None:
            raise AttributeError(
                'Related class must be registered before '
                'it can be queried')

        qs = self.queryset_class(self.destination_class).filter({
            self.foreign_key: getattr(instance, self.primary_key)
        })

        # If we have prefetched related features, populate the QuerySet cache.
        qs._cache = instance._prefetch_cache.get(self.name, None)

        return qs



import inspect
import os
from cuuats.datamodel.features import BaseFeature, relate
from cuuats.datamodel.fields import BlobField, GeometryField, GlobalIDField, \
    OIDField, StringField, NumericField
from cuuats.datamodel.workspaces import WorkspaceManager


def feature_class_factory(path, register=True, exclude=[],
                          follow_relationships=True, class_name=None):
    """
    Create a feature class by introspecting the workspace.
    """

    workspace_path, fc_name = os.path.split(path)
    workspace = WorkspaceManager().get(workspace_path)

    class Feature(BaseFeature):
        pass

    # Set the name of the feature class.
    Feature.__name__ = str(class_name or fc_name)

    # Set the module of the feature class to the module of the caller.
    frame = inspect.stack()[1]
    Feature.__module__ = inspect.getmodule(frame[0])

    # Add fields to the feature class.
    # TODO: Handle Date and Raster field types.
    # TODO: Handle field names that conflict with Feature attributes
    # and methods.
    for (db_name, db_field) in workspace.get_layer_fields(fc_name).items():
        field = {
            'Blob': BlobField,
            'Geometry': GeometryField,
            'Guid': GlobalIDField,
            'OID': OIDField,
            'String': StringField,
        }.get(db_field.type, NumericField)(
            (db_field.aliasName or db_name),
            required=db_field.required)

        setattr(Feature, db_name, field)

    # Set up relationships.
    if follow_relationships:
        for rc_info in workspace.list_relationships(fc_name):
            if rc_info.is_attachment:
                continue

            if rc_info.origin == fc_name:
                related_path = os.path.join(
                    workspace_path, rc_info.destination)
                RelatedFeature = feature_class_factory(
                    related_path, register=False, follow_relationships=False)
                relate(Feature, RelatedFeature, rc_info.primary_key,
                       rc_info.foreign_key)
                if register:
                    RelatedFeature.register(related_path)
            else:
                related_path = os.path.join(workspace_path, rc_info.origin)
                RelatedFeature = feature_class_factory(
                    related_path, register=False, follow_relationships=False)
                relate(RelatedFeature, Feature, rc_info.primary_key,
                       rc_info.foreign_key)
                if register:
                    RelatedFeature.register(related_path)

    # Register the new feature class.
    if register:
        Feature.register(path)

    return Feature



from cuuats.datamodel.workspaces import WorkspaceManager, Workspace
from cuuats.datamodel.fields import BaseField, OIDField, GlobalIDField, \
    GeometryField, StringField, NumericField, CalculatedField, ScaleField, \
    WeightsField, MethodField, ForeignKey
from cuuats.datamodel.features import BaseFeature, require_registration
from cuuats.datamodel.scales import BaseScale, BreaksScale, DictScale, \
    StaticScale, ScaleLevel
from cuuats.datamodel.factory import feature_class_factory
from cuuats.datamodel.query import Q
from cuuats.datamodel.domains import D, CodedValue



import arcpy
import logging
import os
import re
from collections import namedtuple, OrderedDict
from contextlib import contextmanager
from time import time
from cuuats.datamodel.exceptions import ObjectDoesNotExist, \
    MultipleObjectsReturned
from cuuats.datamodel.utils import Singleton


class WorkspaceManager(object):
    """
    Workspace manager for manipulating workspaces.
    """

    __metaclass__ = Singleton

    def __init__(self):
        self.workspaces = {}

    def get(self, path):
        """
        Get a workspace corresponding to the given path.
        """

        if path not in self.workspaces:
            self.workspaces[path] = Workspace(path)
        return self.workspaces[path]

    def clear(self):
        """
        Clear all cached workspaces.
        """

        self.workspaces = {}


class Workspace(object):
    """
    A workspace representing a file geodatabase or SDE.
    """

    RelationshipInfo = namedtuple(
        'RelationshipInfo', ['origin', 'destination', 'primary_key',
                             'foreign_key', 'is_attachment'])

    def __init__(self, path):
        self.path = path
        self.domains = \
            dict([(d.name, d) for d in arcpy.da.ListDomains(self.path)])
        self.editor = arcpy.da.Editor(self.path)

    def list_relationships(self, layer_name):
        """
        List the relationships that this feature class participates in.
        Relationship information is returned as a RelationshipInfo tuple.
        """

        layer = arcpy.Describe(os.path.join(self.path, layer_name))
        return [self.get_relationship_info(rc_name)
                for rc_name in layer.relationshipClassNames]

    def get_relationship_info(self, rc_name):
        """
        Extract the origin, destination, primary and foreign keys from a
        RelationshipClass Describe result or the name of the relationship
        class.
        """

        rc = arcpy.Describe(os.path.join(self.path, rc_name))
        origin = rc.originClassNames[0]
        destination = rc.destinationClassNames[0]
        keys = rc.OriginClassKeys
        primary = [k[0] for k in keys if k[1] == 'OriginPrimary'][0]
        foreign = [k[0] for k in keys if k[1] == 'OriginForeign'][0]
        is_attachment = rc.isAttachmentRelationship
        return self.RelationshipInfo(
            origin, destination, primary, foreign, is_attachment)

    def get_attachment_info(self, layer_name):
        """
        If this layer has attachments, return a named tuple describing the
        attachment relationship.
        """

        for rc_info in self.list_relationships(layer_name):
            if rc_info.is_attachment:
                return rc_info
        return None

    def get_layer_fields(self, layer_name):
        """
        Returns a dictionary of fields for the given layer.
        """

        layer_path = os.path.join(self.path, layer_name)
        return OrderedDict([(f.name, f) for f in arcpy.ListFields(layer_path)])

    def count_rows(self, layer_name, where_clause=None):
        """
        Count the number of rows meeting the given criteria.
        """

        # TODO: Investigate whether using select by attributes would be more
        # efficient.

        return len(list(self.iter_rows(
            layer_name, ['OID@'], where_clause=where_clause)))

    def get_row(self, layer_name, field_names, where_clause=None):
        """
        Retrieve a row from the specified feature class.
        """

        rows_cursors = self.iter_rows(
            layer_name, field_names, where_clause=where_clause, limit=2)
        rows = [r for (r, c) in rows_cursors]

        if len(rows) == 1:
            return rows[0]
        elif len(rows) == 0:
            raise ObjectDoesNotExist(where_clause)
        raise MultipleObjectsReturned(where_clause)

    def iter_rows(self, layer_name, field_names, update=False,
                  where_clause=None, limit=None, prefix=None, postfix=None):
        """
        Iterate over rows of the specified layer.
        """

        layer_path = os.path.join(self.path, layer_name)
        cursor_factory = arcpy.da.SearchCursor

        if update:
            cursor_factory = arcpy.da.UpdateCursor

        logging.debug(
            '{cursor}: SELECT {prefix}{fields} FROM '
            '{table}{where}{postfix}'.format(
                cursor=cursor_factory.__name__,
                prefix=prefix + ' ' if prefix else '',
                fields=', '.join(field_names),
                table=layer_name,
                where=' WHERE ' + where_clause if where_clause else '',
                postfix=' ' + postfix if postfix else ''))

        with cursor_factory(layer_path, field_names, where_clause,
                            sql_clause=(prefix, postfix)) as cursor:
            for row in cursor:

                # Load memoryview data into memory so that we don't lose access
                # to it in the next iteration.
                # TODO: Is there a better way to deal with this issue?
                values = [v.tobytes() if isinstance(v, memoryview) else v
                          for v in row]

                if limit is None or limit > 0:
                    yield (values, cursor)
                    if limit is not None:
                        limit -= 1
                else:
                    break

    def update_row(self, cursor, values):
        """
        Update the active row in the current cursor with the given values.
        """

        if not isinstance(cursor, arcpy.da.UpdateCursor):
            raise TypeError('Invalid cursor')
        cursor.updateRow(values)

    def insert_row(self, layer_name, field_names, values):
        """
        Insert a row in the table.
        """

        layer_path = os.path.join(self.path, layer_name)
        self.editor.startEditing(False, False)

        with arcpy.da.InsertCursor(layer_path, field_names) as cursor:
            oid = cursor.insertRow(values)

        self.editor.stopEditing(True)
        return oid

    @contextmanager
    def edit(self, versioned=True):
        self.editor.startEditing(False, versioned)
        if versioned:
            self.editor.startOperation()
        try:
            yield
            if versioned:
                self.editor.stopOperation()
            self.editor.stopEditing(True)
        except Exception, e:
            if versioned:
                self.editor.abortOperation()
            self.editor.stopEditing(False)
            raise e

    def get_domain(self, domain_name, domain_type=None):
        """
        Get the named domain if it exists.
        """

        domain = self.domains.get(domain_name, None)

        if domain is None:
            raise NameError('Invalid domain name: %s', (domain_name,))

        if domain_type is not None and domain.domainType != domain_type:
            raise TypeError(
                '%s is not a domain of type %s' % (domain_name, domain_type))

        return domain

    def get_coded_value(self, domain_name, description):
        """
        Get the coded value from a domain using the description.
        """

        domain = self.get_domain(domain_name, 'CodedValue')

        for (code, desc) in domain.codedValues.items():
            if desc == description:
                return code

        raise ValueError('Domain %s has no code for description %s' %
                         (domain_name, description))

    def add_field(self, layer_name, field_name, storage):
        """
        Add a field to a feature class.
        """

        if 'field_type' not in storage:
            raise KeyError('The storage dictionary must contain the '
                           'key field_type.')

        layer_path = os.path.join(self.path, layer_name)
        arcpy.AddField_management(
            layer_path,
            field_name,
            **storage)

    def _make_layer_name(self):
        """
        Returns a unique layer name.
        """

        return 'layer_%s' % (str(time()).replace('.', ''),)

    @contextmanager
    def make_layer(self, fc_name, where_clause=None):
        """
        Temporarily make a feature class into a layer so that it can be
        used with geoprocessing tools that require a layer.
        """

        layer_name = self._make_layer_name()
        fc_path = os.path.join(self.path, fc_name)

        arcpy.MakeFeatureLayer_management(fc_path, layer_name)
        yield layer_name

        arcpy.Delete_management(layer_name)

    def summarize(self, fc_name, fields, where_clause=None):
        """
        Generate summary statistics for a feature class.
        """

        summary_name = self._make_layer_name()
        summary_path = 'in_memory/%s' % (summary_name,)

        with self.make_layer(fc_name) as layer_name:
            if where_clause:
                arcpy.SelectLayerByAttribute_management(
                    layer_name, where_clause=where_clause)

            arcpy.Statistics_analysis(
                layer_name, summary_path, fields)

        summary_fields = [re.sub(r'[^A-Za-z0-9]', '_', '%s_%s' % (s, f))
                          for (f, s) in fields]

        for row in arcpy.da.SearchCursor(summary_path, summary_fields):
            return dict(zip(summary_fields, row))

    def update_spatial_relationship(self, rc_name, rel_type='INTERSECT',
                                    search_radius=None):
        """
        Assign or update values for a foreign key based on the spatial
        relationship of the features.
        """

        # TODO: Move this method to a separate class dedicated to dealing
        # with spatial relationships.

        rc_info = self.get_relationship_info(rc_name)
        dest_path = os.path.join(self.path, rc_info.destination)
        origin_path = os.path.join(self.path, rc_info.origin)

        # Create field mappings. We really don't want any fields
        # included, so we add the origin primary key, which will
        # be removed automatically.
        field_mappings = arcpy.FieldMappings()
        field_map = arcpy.FieldMap()
        field_map.addInputField(origin_path, rc_info.primary_key)
        field_mappings.addFieldMap(field_map)

        # Execute the spatial join.
        join_name = self._make_layer_name()
        join_path = 'in_memory/%s' % (join_name,)
        arcpy.SpatialJoin_analysis(
            dest_path, origin_path, join_path, 'JOIN_ONE_TO_MANY',
            'KEEP_COMMON', field_mappings, rel_type, search_radius)

        # Create a mapping: destination -> origin.
        with arcpy.da.SearchCursor(
                join_path, ['TARGET_FID', 'JOIN_FID']) as cursor:
            oid_map = dict([(str(row[0]), row[1]) for row in cursor])

        # Delete the join layer.
        arcpy.Delete_management(join_path)
        if not oid_map:
            return

        # Update the forien key.
        for (row, cursor) in self.iter_rows(
                rc_info.destination, ['OID@', rc_info.foreign_key], True):
            oid, fk = row
            new_fk = oid_map.get(str(oid), None)
            if fk != new_fk:
                self.update_row(cursor, [oid, new_fk])



"""
Utility functions and classes.
"""


class Singleton(type):
    """
    Metaclass for creating a singleton class. Source:
    http://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args, **kwargs)
        return cls._instances[cls]



# -*- coding: utf-8 -*-

from setuptools import find_packages
from setuptools import setup

version = '0.1.1'

setup(name='cuuats.datamodel',
      version=version,
      description='A lightweight data access layer for ArcGIS',
      long_description='\n'.join([open(f).read() for f in [
          'README.rst',
          'HISTORY.rst'
      ]]),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Win32 (MS Windows)',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Topic :: Database',
          'Topic :: Scientific/Engineering :: GIS',
          'Programming Language :: Python :: 2.7',
      ],
      keywords='ArcGIS GIS data geodatabase',
      author='Matt Yoder',
      author_email='myoder@ccrpc.org',
      url='https://cuuats.org/',
      download_url='https://github.com/CUUATS/cuuats.datamodel/tarball/0.1.1',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['cuuats']
      )



class DeferredValue(object):
    """
    A field value that is only retrieved from the database when needed.
    """

    def __init__(self, field_name, db_name):
        self.field_name = field_name
        self.db_name = db_name



import os
import re
from collections import OrderedDict, namedtuple
from cuuats.datamodel.fields import BaseField, OIDField, CalculatedField, \
    ForeignKey, NumericField, StringField, BlobField, GeometryField
from cuuats.datamodel.field_values import DeferredValue
from cuuats.datamodel.query import Q, Manager, SQLCompiler
from cuuats.datamodel.workspaces import WorkspaceManager

IDENTIFIER_RE = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*')


def require_registration(fn):
    """
    Decorator to check that the class has been registered with a workspace.
    """

    def wrapper(*args, **kwargs):
        if getattr(args[0], 'workspace', None) is None:
            raise AttributeError(
                'Feature class %s must be registered' %
                (args[0].__name__,))
        return fn(*args, **kwargs)

    return wrapper


class FieldSet(OrderedDict):

    def __init__(self, *args, **kwargs):
        super(FieldSet, self).__init__(*args, **kwargs)
        self.oid_field = None
        self.geom_field = None
        # Set the name and db_name of fields. This also happens during
        # field registration, but we may need to access these attributes
        # before the field is registered.
        for field_name, field in self.items():
            field.name = field_name
            field.db_name = field.db_name or field_name
            if isinstance(field, OIDField):
                self.oid_field = field
            elif isinstance(field, GeometryField):
                self.geom_field = field

    def get_name(self, db_name):
        """
        Given the database name, get the corresponding field name.
        """

        for (field_name, field) in self.items():
            if (field.db_name or field_name) == db_name:
                return field_name

    def get_db_name(self, field_name):
        """
        Get the database name for a field.
        """

        return self[field_name].db_name


class FieldManager(object):

    def __get__(self, instance, owner):
        if getattr(owner, '_fields', None) is None:
            self._cache_fields(owner)
        return owner._fields

    def _cache_fields(self, owner):
        fields = []

        # Walk the inheritance chain looking for fields.
        for subcls in type.mro(owner):
            fields.extend([(f.order, f.creation_index, n, f) for (n, f) in
                          subcls.__dict__.items()
                          if isinstance(f, BaseField)])
        # Cache the resutling FieldSet.
        owner._fields = FieldSet([d[2:] for d in sorted(fields)])


DiffValues = namedtuple('DiffValues', ['old', 'new'])


class BaseFeature(object):
    """
    Base class used to interact with data stored in a geodatabase feature
    class.
    """

    name = None
    workspace = None
    attachments = None
    db_values = None
    objects = Manager()
    fields = FieldManager()
    related_classes = None

    @classmethod
    def register(cls, path):
        """
        Connect this feature class to a feature class in a workspace.
        """

        workspace_path, cls.name = os.path.split(path)
        cls.workspace = WorkspaceManager().get(workspace_path)

        if not issubclass(cls, BaseAttachment):
            attachment_info = cls.workspace.get_attachment_info(cls.name)
            if attachment_info is not None:
                Attachment = attachment_class_factory(
                    cls, attachment_info.primary_key,
                    attachment_info.foreign_key)
                Attachment.register(
                    os.path.join(workspace_path, attachment_info.destination))
                cls.attachment_class = Attachment

        # Register fields with the workspace
        layer_fields = cls.workspace.get_layer_fields(cls.name)
        for (field_name, field) in cls.fields.items():
            field.register(
                cls.workspace, cls, field_name, cls.name, layer_fields)

    @classmethod
    @require_registration
    def count(cls, where_clause=None):
        """
        Count the number of features matching the given where clause.
        """

        return cls.workspace.count_rows(cls.name, where_clause)

    @classmethod
    @require_registration
    def sync_fields(cls, modify=False, remove=False):
        """
        Adds (and optionally, modify or removes) database fields to match
        those defined in this feature class.
        """

        if modify or remove:
            raise NotImplementedError('Modification and removal of fields '
                                      'are not yet supported')

        layer_fields = cls.workspace.get_layer_fields(cls.name)
        added_fields = []

        for (field_name, field) in cls.fields.items():
            if field.db_name not in layer_fields and field.storage:
                storage = field.storage
                if 'field_alias' not in storage:
                    storage['field_alias'] = field.label
                cls.workspace.add_field(cls.name, field_name, storage)
                added_fields.append(field_name)

        # Reregister fields
        layer_fields = cls.workspace.get_layer_fields(cls.name)
        for (field_name, field) in cls.fields.items():
            field.register(
                cls.workspace, cls, field_name, cls.name, layer_fields)

    @require_registration
    def __init__(self, **kwargs):
        for field_name in kwargs.keys():
            if field_name not in self.fields.keys():
                raise KeyError('Invalid field name: %s' % (field_name))

        self.values = kwargs
        self.db_values = dict(
            [(self.fields[k].db_name, v) for (k, v) in kwargs.items()
             if not isinstance(v, DeferredValue)])
        self._prefetch_cache = {}

    def __repr__(self):
        name = self.name or '(unregistered)'
        return '<%s: %s>' % (self.__class__.__name__, name)

    @property
    def oid(self):
        """
        Returns the object ID (primary key) for this feature.
        """

        return getattr(self, self.fields.oid_field.name)

    @property
    def oid_where(self):
        """
        Where clause for selecting this feature.
        """

        compiler = SQLCompiler(self.__class__)
        return compiler.compile(Q({self.fields.oid_field.name: self.oid}))

    def get_deferred_values(self, fields=[]):
        """
        Retrieve deferred values from the database.
        """

        if not fields:
            fields = dict([(v.field_name, v.db_name) for v in
                           self.values.values()
                           if isinstance(v, DeferredValue)])

        values = self.workspace.get_row(
            self.name, fields.values(), self.oid_where)

        self.values.update(dict(zip(fields.keys(), values)))

    def clean(self):
        """
        Perform cleaning of the raw data.
        """
        # Overridden by subclasses.
        pass

    def validate(self):
        """
        Perform validation on each field in the feature, and return any
        validation error messages. Deferred values that have not been
        retrieved are skipped, as are foreign keys and calculated fields.
        """

        messages = []
        for (field_name, field) in self.fields.items():
            if isinstance(field, (ForeignKey, CalculatedField)):
                continue

            # Skip deferred values that have not been retrieved.
            if isinstance(self.values.get(field_name, None), DeferredValue):
                continue

            value = getattr(self, field_name)
            if value is None and self.check_condition(
                    field.required_if, default=False):
                messages.append('%s is missing' % (field.label,))
            else:
                messages.extend(field.validate(value))
        return messages

    def serialize(self):
        """
        Return a list of values for the fields in this feature. Deferred values
        that have not been retrieved are excluded.
        """

        return dict(
            [(f.db_name, self.values.get(n, None))
             if isinstance(f, ForeignKey) else
             (f.db_name, getattr(self, n)) for (n, f) in self.fields.items()
             if not isinstance(self.values.get(n), DeferredValue)])

    def diff(self):
        """
        Returns a dictionary containing unsaved changes.
        """

        new = self.serialize()
        db_fields = [(f.db_name, f) for f in self.fields.values()]
        return dict(
            [(n, DiffValues(self.db_values[n], new[n]))
             for (n, f) in db_fields
             if n in new and n in self.db_values and
             f.has_changed(self.db_values[n], new[n])])

    def save(self):
        """
        Update the corresponding row in feature class.
        """

        oid = getattr(self, self.fields.oid_field.name)
        if oid is None:
            return self._insert(self.serialize())
        else:
            changes = self.diff()
            if not changes:
                return False

            field_values = dict([(n, v.new) for (n, v) in changes.items()])
            return self._update(oid, field_values)

    def _insert(self, field_values):
        # Remove the (null) OID from the fields and values.
        del field_values[self.fields.oid_field.db_name]
        field_names, values = zip(*field_values.items())

        # Perform the insert.
        oid = self.workspace.insert_row(self.name, field_names, values)

        # Set the OID.
        setattr(self, self.fields.oid_field.name, oid)

        # Update the db_row.
        field_values[self.fields.oid_field.db_name] = oid
        self.db_values.update(field_values)

        return True

    def _update(self, oid, field_values):
        where_clause = '%s = %i' % (self.fields.oid_field.name, oid)
        field_names, values = zip(*field_values.items())
        updated_count = 0
        for (row, cursor) in self.workspace.iter_rows(
                self.name, field_names, True, where_clause):
            self.workspace.update_row(cursor, values)
            updated_count += 1

        if updated_count == 0:
            raise LookupError('A row with OID %i was not found' % (oid,))
        else:
            self.db_values.update(field_values)
            return True

    def eval(self, expression):
        """
        Evaluate the expression in the context of the feature instance.
        """

        identifiers = IDENTIFIER_RE.findall(expression)
        # Limit retrieval of field values to field names that are found in the
        # expression in order to prevent recursion for calculated fields.
        locals_dict = dict([(f, getattr(self, f)) for f
                            in self.fields.keys() if f in identifiers])
        locals_dict.update({'self': self})
        return eval(expression, {}, locals_dict)

    def check_condition(self, condition, default=True):
        """
        Returns a boolean indicating whether the condition is true for the
        this feature instance.
        """

        if condition is None:
            return default

        return bool(self.eval(condition))


def relate(origin_class, destination_class, primary_key, foreign_key,
           fk_field_name=None, fk_label=None, related_name=None):
    """
    Establish a relationship between two classes by assigning a ForeignKey
    to the destination class.
    """

    fk_field_name = fk_field_name or foreign_key
    fk_label = fk_label or fk_field_name
    pk_field_name = origin_class.fields.get_name(primary_key)

    field = ForeignKey(
        fk_label,
        db_name=foreign_key,
        origin_class=origin_class,
        primary_key=pk_field_name,
        related_name=related_name)

    setattr(destination_class, fk_field_name, field)


class BaseAttachment(BaseFeature):
    """
    A file attached to an ArcGIS feature class.
    """

    attachment_id = OIDField(
        'Attachment ID',
        db_name='ATTACHMENTID')

    file_name = StringField(
        'File Name',
        db_name='ATT_NAME')

    content_type = StringField(
        'Content Type',
        db_name='CONTENT_TYPE')

    file_size = NumericField(
        'File Size',
        db_name='DATA_SIZE')

    data = BlobField(
        'File Data',
        db_name='DATA')

    @property
    def has_data(self):
        """
        Does this attachment have data?
        """

        return self.data is not None

    def save_to(self, path, filename=None, overwrite=False, data=None):
        """
        Save the attached file to the specified directory path.
        """

        if not self.has_data and data is None:
            raise ValueError('Cannot save an attachment with no data')

        if data is None:
            data = self.data

        file_path = os.path.join(path, filename or self.file_name)
        if os.path.isfile(file_path) and not overwrite:
            raise IOError('%s already exists' % (file_path,))

        with open(file_path, 'wb') as file:
            file.write(data)


def attachment_class_factory(
        origin_class, primary_key, foreign_key, related_name='attachments',
        base_class=BaseAttachment):
    """
    Each feature class needs a unique attachment class so that it can be
    registered for the appropriate attachment table in the workspace.
    """

    class Attachment(base_class):
        """
        A file attachment.
        """

    relate(origin_class, Attachment, primary_key, foreign_key, 'feature',
           'Feature', related_name)

    return Attachment



import arcpy
import os
import shutil
import tempfile
import unittest
from cuuats.datamodel.exceptions import ObjectDoesNotExist, \
    MultipleObjectsReturned
from cuuats.datamodel.workspaces import Workspace
from cuuats.datamodel.fields import BaseField, OIDField, GeometryField, \
    StringField, NumericField, ScaleField, MethodField, WeightsField
from cuuats.datamodel.features import BaseFeature
from cuuats.datamodel.scales import BreaksScale, DictScale
from cuuats.datamodel.domains import CodedValue, D
from cuuats.datamodel.workspaces import WorkspaceManager


def setUpModule():
    WorkspaceFixture.setUpModule()


def tearDownModule():
    WorkspaceFixture.tearDownModule()


class WorkspaceFixture(object):

    GDB_NAME = 'test.gdb'
    GDB_BACKUP_NAME = 'test_backup.gdb'
    DOMAIN_NAME = 'YesOrNo'
    DOMAIN_DESCRIPTION = 'Domain for yes or no responses'
    DOMAIN_FIELD_TYPE = 'SHORT'
    DOMAIN_TYPE = 'CODED'
    DOMAIN_VALUES = {
        50: 'No',
        100: 'Yes',
        101: 'N/A',
    }
    FEATURE_CLASS_NAME = 'Widget'
    FEATURE_CLASS_TYPE = 'POINT'
    FEATURE_CLASS_FIELDS = (
        ('widget_name', 'TEXT', 100, None),
        ('widget_description', 'TEXT', None, None),
        ('widget_number', 'LONG', None, None),
        ('widget_available', DOMAIN_FIELD_TYPE, None, DOMAIN_NAME),
        ('widget_price', 'FLOAT', None, None),
        ('widget_number_score', 'DOUBLE', None, None),
    )
    FEATURE_CLASS_DATA = (
        ('Widget A+ Awesome', None, 12345, 100, 10.50, None, (2.5, 3.0)),
        ('B-Widgety Widget', None, None, 50, None, None, (-2.0, 5.5)),
        ('My Widget C', 'Best widget', None, None, None, None, (0.0, 4.0)),
    )

    @classmethod
    def setUpModule(cls):
        # Create a new file geodatabase.
        cls.workspace_dir = tempfile.mkdtemp()
        cls.gdb_backup_path = os.path.join(
            cls.workspace_dir, cls.GDB_BACKUP_NAME)
        arcpy.CreateFileGDB_management(cls.workspace_dir, cls.GDB_BACKUP_NAME)

        # Add a coded values domain, and populate it with coded values.
        arcpy.CreateDomain_management(
            cls.gdb_backup_path, cls.DOMAIN_NAME, cls.DOMAIN_DESCRIPTION,
            cls.DOMAIN_FIELD_TYPE, cls.DOMAIN_TYPE)

        for (code, desc) in cls.DOMAIN_VALUES.items():
            arcpy.AddCodedValueToDomain_management(
                cls.gdb_backup_path, cls.DOMAIN_NAME, code, desc)

        # Create a feature class.
        fc_path = os.path.join(cls.gdb_backup_path, cls.FEATURE_CLASS_NAME)
        arcpy.CreateFeatureclass_management(
            cls.gdb_backup_path, cls.FEATURE_CLASS_NAME,
            cls.FEATURE_CLASS_TYPE)

        # Add some fields to the feature class.
        for (field_name, field_type, field_precision, field_domain) in \
                cls.FEATURE_CLASS_FIELDS:
            arcpy.AddField_management(
                fc_path, field_name, field_type, field_precision,
                field_domain=field_domain)

        # Populate the feature class with data.
        field_names = [f[0] for f in cls.FEATURE_CLASS_FIELDS] + ['SHAPE@XY']
        with arcpy.da.InsertCursor(fc_path, field_names) as cursor:
            for row in cls.FEATURE_CLASS_DATA:
                cursor.insertRow(row)

    @classmethod
    def tearDownModule(cls):
        shutil.rmtree(cls.workspace_dir)

    def setUp(self):
        # Restore the geodatabase from backup.
        self.gdb_path = os.path.join(self.workspace_dir, self.GDB_NAME)
        shutil.copytree(self.gdb_backup_path, self.gdb_path)

        # Create a workspace from the GDB.
        self.fc_path = os.path.join(self.gdb_path, self.FEATURE_CLASS_NAME)
        self.workspace = Workspace(self.gdb_path)

        # Create a feature matching the workspace.
        class Widget(BaseFeature):
            """
            Test feature class.
            """

            OBJECTID = OIDField('OID')
            widget_name = StringField('Widget Name', required=True)
            widget_description = StringField('Widget Description')
            widget_number = NumericField('Widget Number', required=True)
            widget_available = NumericField('Is Widget Available?',
                                            required=True)
            widget_price = NumericField('Widget Price',
                                        required_if='self.is_available')
            widget_number_score = ScaleField(
                'Widget Number Score',
                scale=BreaksScale([100, 500, 1000], [1, 2, 3, 4]),
                value_field='widget_number',
                storage={'field_type': 'DOUBLE'}
            )
            Shape = GeometryField('Shape')

            @property
            def is_available(self):
                return self.widget_available == D('Yes')

        self.cls = Widget

    def tearDown(self):
        if self.cls.workspace is not None:
            del self.cls.workspace
        del self.cls
        del self.workspace
        WorkspaceManager().clear()
        shutil.rmtree(self.gdb_path)


class TestFields(unittest.TestCase):

    def setUp(self):
        class MyFeatureClass(object):
            base_field = BaseField(
                'Base Field',
                choices=[0, 1, 2],
                required=True)
            oid_field = OIDField('OBJECTID')
            geometry_field = GeometryField('SHAPE')
            string_field = StringField('String Field', required=True)
            numeric_field = NumericField(
                'Numeric Field', min=0, max=10, db_scale=3)
            double_method = MethodField(
                'Numeric Field Doubled', method_name='_double')
            weights_field = WeightsField('Weights Field', weights={
                'numeric_field': 0.25,
                'double_method': 0.75
            }, default=0)

            def __init__(self):
                self.values = {}

            def _double(self, field_name):
                if self.numeric_field is None:
                    return None
                return self.numeric_field * 2

            def get_field(self, field_name):
                return self.__class__.__dict__[field_name]

            def check_condition(self, condition):
                return True

        self.inst_a = MyFeatureClass()
        self.inst_b = MyFeatureClass()

    def test_base_field_get_set(self):
        self.assertTrue(
            isinstance(self.inst_a.get_field('base_field'), BaseField),
            'base field is not an instance of BaseField')

        self.assertEqual(self.inst_a.base_field, None)

        self.inst_a.base_field = 100
        self.assertEqual(self.inst_a.base_field, 100)
        self.assertNotEqual(self.inst_a.base_field, self.inst_b.base_field)

        self.inst_b.base_field = 200
        self.assertEqual(self.inst_a.base_field, 100)
        self.assertEqual(self.inst_b.base_field, 200)

    def test_base_field_validation_required(self):
        field = self.inst_a.get_field('base_field')
        self.assertTrue('Base Field is missing' in field.validate(None),
                        'validation message for required field is missing')
        self.assertEqual(len(field.validate(1)), 0)

    def test_base_field_validation_choices(self):
        field = self.inst_a.get_field('base_field')
        self.assertTrue('Base Field is invalid' in field.validate(10),
                        'no validation message for coded value domain')
        self.assertEqual(len(field.validate(2)), 0)

    def test_string_field_validation(self):
        field = self.inst_a.get_field('string_field')
        self.assertTrue('String Field is missing' in field.validate(''),
                        'no validation message for empty string')

    def test_numeric_field_validation(self):
        field = self.inst_a.get_field('numeric_field')
        self.assertTrue('Numeric Field out of range' in field.validate(25),
                        'no validation message for value above field max')
        self.assertTrue('Numeric Field out of range' in field.validate(-1),
                        'no validation message for value below field min')
        self.assertEqual(len(field.validate(0)), 0)
        self.assertEqual(len(field.validate(5)), 0)
        self.assertEqual(len(field.validate(10)), 0)

    def test_numeric_field_has_changed(self):
        field = self.inst_a.get_field('numeric_field')
        self.assertFalse(field.has_changed(10, 10))
        self.assertTrue(field.has_changed(10, 15))
        self.assertFalse(field.has_changed(1.002, 1.0022))
        self.assertTrue(field.has_changed(1.002, 1.0026))

    def test_method_field(self):
        self.inst_a.numeric_field = 2
        self.inst_b.numeric_field = 3
        self.assertEqual(self.inst_a.double_method, 4)
        self.assertEqual(self.inst_b.double_method, 6)

    def test_weights_field(self):
        self.assertEqual(self.inst_a.weights_field, 0)
        self.inst_a.numeric_field = 2
        self.assertEqual(self.inst_a.weights_field, 3.5)


class TestWorkspace(WorkspaceFixture, unittest.TestCase):

    def test_get_attachment_info(self):
        no_attach = self.workspace.get_attachment_info(self.FEATURE_CLASS_NAME)
        self.assertEqual(no_attach, None)

        arcpy.EnableAttachments_management(self.fc_path)
        attach = self.workspace.get_attachment_info(self.FEATURE_CLASS_NAME)
        self.assertEqual(attach.origin, self.FEATURE_CLASS_NAME)

    def test_get_layer_fields(self):
        layer_fields = self.workspace.get_layer_fields(self.FEATURE_CLASS_NAME)
        for field_def in self.FEATURE_CLASS_FIELDS:
            field_name = field_def[0]
            self.assertTrue(field_name in layer_fields,
                            'missing layer field: %s' % (field_name,))

    def test_count_rows(self):
        count = self.workspace.count_rows(
            self.FEATURE_CLASS_NAME, 'widget_number = 12345')
        self.assertEqual(count, 1)

    def iter_rows(self):
        rows = list(self.workspace.iter_rows(
            self.FEATURE_CLASS_NAME,
            ['widget_name', 'widget_available'],
            where_clause='widget_number = 12345'))

        self.assertEqual(
            len(rows), 1, 'where clause does not filter results')

        self.asserEqual(
            ['Widget A+ Awesome', 100], rows[0], 'incorrect row values')

    def test_update_row(self):
        field_names = [f[0] for f in self.FEATURE_CLASS_FIELDS]

        with self.workspace.edit():
            for (row, cursor) in self.workspace.iter_rows(
                    self.FEATURE_CLASS_NAME, field_names, update=True):
                row[0] = row[0].replace('Widget', 'Foo')
                self.workspace.update_row(cursor, row)

        with arcpy.da.SearchCursor(self.fc_path, ['widget_name']) as cursor:
            for feature in cursor:
                widget_name = feature[0]
                self.assertTrue('Widget' not in widget_name)
                self.assertTrue('Foo' in widget_name)

    def test_insert_row(self):
        field_names = [f[0] for f in self.FEATURE_CLASS_FIELDS]
        values = ('DWIDGET', 'D-Widget', None, None, None, None)
        self.workspace.insert_row(self.FEATURE_CLASS_NAME, field_names, values)

        with arcpy.da.SearchCursor(self.fc_path, ['widget_name']) as cursor:
            widget_names = [row[0] for row in cursor]
        self.assertTrue('DWIDGET' in widget_names)

    def test_domains(self):
        self.assertEqual(len(self.workspace.domains), 1,
                         'incorrect number of domains in workspace')
        self.assertTrue(self.DOMAIN_NAME in self.workspace.domains,
                        'domain is missing from the workspace domains')

    def test_get_domain(self):
        with self.assertRaises(NameError):
            self.workspace.get_domain('NotADomain')

        with self.assertRaises(TypeError):
            self.workspace.get_domain(self.DOMAIN_NAME, 'Range')

        domain = self.workspace.get_domain(self.DOMAIN_NAME, 'CodedValue')
        self.assertTrue(isinstance(domain, arcpy.da.Domain))

    def test_get_coded_value(self):
        self.assertEqual(
            self.workspace.get_coded_value(self.DOMAIN_NAME, 'No'), 50,
            'incorrect domain code')

        with self.assertRaises(ValueError):
            self.workspace.get_coded_value(self.DOMAIN_NAME, 'NotADescription')

    def test_add_field(self):
        num_fields = len(self.workspace.get_layer_fields(
            self.FEATURE_CLASS_NAME))

        with self.assertRaises(KeyError):
            self.workspace.add_field(
                self.FEATURE_CLASS_NAME,
                'widget_color',
                {})

        self.workspace.add_field(
            self.FEATURE_CLASS_NAME,
            'widget_color',
            {'field_type': 'TEXT', 'field_length': 100})

        fields = self.workspace.get_layer_fields(self.FEATURE_CLASS_NAME)
        self.assertEqual(len(fields), num_fields + 1,
                         'Widget color field not added')
        self.assertTrue('widget_color' in fields.keys())


class TestRegisterFeature(WorkspaceFixture, unittest.TestCase):

    def test_register(self):
        with self.assertRaises(AttributeError):
            self.cls([])

        with self.assertRaises(AttributeError):
            self.cls.iter()

        self.cls.register(self.fc_path)
        self.assertEqual(self.cls.workspace.path, self.workspace.path)
        for (field_name, field) in self.cls.fields.items():
            self.assertEqual(field.name, field_name,
                             'field name is not set correctly')
        self.assertEqual(
            len(self.cls.__dict__['widget_available'].choices), 3,
            'choices from domain is not assigned correctly')


class TestFeature(WorkspaceFixture, unittest.TestCase):

    def setUp(self):
        super(TestFeature, self).setUp()
        self.cls.register(self.fc_path)
        self.instance = self.cls(
            OBJECTID=1, widget_name='My Widget', widget_number=300,
            widget_available=50, Shape=None)

    def test_get_fields(self):
        field_names = ['OBJECTID', 'widget_name', 'widget_description',
                       'widget_number', 'widget_available', 'widget_price',
                       'widget_number_score', 'Shape']
        self.assertEqual(self.cls.fields.keys(), field_names)

        del self.cls.widget_name
        self.cls.widget_name = StringField('Widget Name', required=True)
        self.cls._fields = None
        self.assertEqual(
            self.cls.fields.keys()[-1], 'widget_name',
            'fields are not ordered based on creation_index')

        del self.cls.widget_name
        self.cls.widget_name = StringField(
            'Widget Name', required=True, order=-1)
        self.cls._fields = None
        self.assertEqual(
            self.cls.fields.keys()[0], 'widget_name',
            'fields are not ordered base on order argument')

    def test_save_update(self):
        feature = self.cls.objects.get(OBJECTID=1)
        feature.widget_name = 'Some Widget'
        result = feature.save()

        self.assertTrue(result)
        self.assertEqual(self.cls.objects.first().widget_name, 'Some Widget')

    def test_save_insert(self):
        feature_count = self.cls.objects.count()
        feature = self.cls(widget_name='Newest Widget')
        feature.save()

        self.assertEqual(self.cls.objects.count(), feature_count + 1)
        self.assertEqual(self.cls.objects.last().widget_name, 'Newest Widget')

    def test_save_no_change(self):
        # Update all features to set calculated field values.
        for feature in self.cls.objects.all():
            feature.save()

        # Try updating features without changing anything.
        update_count = 0
        for feature in self.cls.objects.all():
            update_count += int(feature.save())
        self.assertEqual(update_count, 0, 'Features updated unnecessarily')

        # Change a value, and check the update count.
        update_count = 0
        for feature in self.cls.objects.all():
            feature.widget_number = 500
            update_count += int(feature.save())
        self.assertEqual(update_count, 3, 'Features not updated after change')

    def test_get_field(self):
        self.assertEqual(self.instance.fields.get('NotAField'), None)
        self.assertTrue(
            isinstance(self.instance.fields.get('widget_name'), StringField))

    def test_set_by_description(self):
        self.instance.widget_available = D('Yes')
        self.assertEqual(self.instance.widget_available, 100)

    def test_get_description_for(self):
        self.assertEqual(
            self.instance.widget_available, D('No'))

    def test_field_value_changed(self):
        self.assertEqual(self.instance.widget_number_score, 2)
        self.instance.widget_number = 800
        self.assertEqual(
            self.instance.widget_number_score, 3,
            'Calculated field not updated')

    def test_sync_fields(self):
        self.assertTrue('widget_number_score' in
                        self.workspace.get_layer_fields(
                            self.FEATURE_CLASS_NAME))

        arcpy.DeleteField_management(self.fc_path, 'widget_number_score')

        self.assertTrue('widget_number_score' not in
                        self.workspace.get_layer_fields(
                            self.FEATURE_CLASS_NAME))

        self.cls.sync_fields()
        self.assertTrue('widget_number_score' in
                        self.workspace.get_layer_fields(
                            self.FEATURE_CLASS_NAME),
                        'Syncing fields did not create widget_number_score')

    def test_required_if(self):
        price_msg = 'Widget Price is missing'
        description_msg = 'Widget Description is missing'

        self.assertTrue(description_msg not in self.instance.validate(),
                        'Optional field produces missing validation message')

        self.assertTrue(price_msg not in self.instance.validate(),
                        'required_if validation message incorrectly generated')

        self.instance.widget_available = D('Yes')
        self.assertTrue(price_msg in self.instance.validate(),
                        'required_if validation message not generated')

        self.instance.widget_price = 20.00
        self.assertTrue(price_msg not in self.instance.validate(),
                        'required_if validation message incorrectly generated')

    def test_diff(self):
        feature = self.cls.objects.get(OBJECTID=1)
        feature.save()
        original_name = feature.widget_name
        feature.widget_name = 'Some Widget'
        self.assertEqual(
            feature.diff(), {'widget_name': (original_name, 'Some Widget')})

        feature.save()
        self.assertEqual(feature.diff(), {})


class TestQuerySet(WorkspaceFixture, unittest.TestCase):

        def setUp(self):
            super(TestQuerySet, self).setUp()
            self.cls.register(self.fc_path)

        def test_all(self):
            widget_names = [row[0] for row in self.FEATURE_CLASS_DATA]
            for feature in self.cls.objects.all():
                self.assertTrue(isinstance(feature, self.cls))
                self.assertTrue(feature.widget_name in widget_names)

        def test_get(self):
            feature = self.cls.objects.get(OBJECTID=2)
            self.assertEqual(feature.OBJECTID, 2)
            self.assertEqual(feature.widget_name, 'B-Widgety Widget')

            with self.assertRaises(ObjectDoesNotExist):
                self.cls.objects.get(OBJECTID=5)

            with self.assertRaises(MultipleObjectsReturned):
                self.cls.objects.get(widget_description=None)

        def test_get_save(self):
            inst_a = self.cls.objects.get(OBJECTID=1)
            inst_b = self.cls.objects.get(OBJECTID=2)

            self.assertEqual(inst_a.widget_number, 12345)
            self.assertEqual(inst_b.widget_number, None)

            inst_a.widget_number = 10
            inst_a.save()

            inst_b.widget_number = 20
            inst_b.save()

            del inst_a
            del inst_b

            widget_numbers = [r[0] for (r, c) in self.workspace.iter_rows(
                self.FEATURE_CLASS_NAME, ['widget_number'])]
            self.assertTrue(10 in widget_numbers)
            self.assertTrue(20 in widget_numbers)


class TestBreaksScale(unittest.TestCase):

    def setUp(self):
        self.breaks_right = BreaksScale([5, 10, 15, 20], [1, 2, 3, 4, 5])
        self.breaks_left = BreaksScale([5, 10, 15, 20], [1, 2, 3, 4, 5], False)

    def test_init(self):
        with self.assertRaises(ValueError):
            BreaksScale([5, 20, 15], [4, 3, 2, 1])

        with self.assertRaises(IndexError):
            BreaksScale([5, 10, 15], [3, 2, 1])

    def test_score_right(self):
        self.assertEqual(self.breaks_right.score(-10), 1)
        self.assertEqual(self.breaks_right.score(5), 1)
        self.assertEqual(self.breaks_right.score(6), 2)
        self.assertEqual(self.breaks_right.score(20), 4)
        self.assertEqual(self.breaks_right.score(100), 5)

    def test_score_left(self):
        self.assertEqual(self.breaks_left.score(-10), 1)
        self.assertEqual(self.breaks_left.score(5), 2)
        self.assertEqual(self.breaks_left.score(6), 2)
        self.assertEqual(self.breaks_left.score(19), 4)
        self.assertEqual(self.breaks_left.score(20), 5)
        self.assertEqual(self.breaks_left.score(100), 5)


class TestDictScale(unittest.TestCase):

    def setUp(self):
        self.scale = DictScale({
            'one': 1,
            'two': 2,
            'three': 3,
        }, 0)

    def test_score(self):
        self.assertEqual(self.scale.score('one'), 1)
        self.assertEqual(self.scale.score('three'), 3)
        self.assertEqual(self.scale.score('notakey'), 0)


class TestCodedValue(unittest.TestCase):

    def test_coded_value_description(self):
        cv = CodedValue(3, 'Bar')
        self.assertEqual(cv.description, 'Bar')

    def test_int_coded_value(self):
        cv = CodedValue(10, 'Integer')
        self.assertTrue(isinstance(cv, CodedValue))
        self.assertTrue(isinstance(cv, int))

    def test_long_coded_value(self):
        cv = CodedValue(10L, 'Long Integer')
        self.assertTrue(isinstance(cv, CodedValue))
        self.assertTrue(isinstance(cv, long))

    def test_float_coded_value(self):
        cv = CodedValue(10.2, 'Floating Point')
        self.assertTrue(isinstance(cv, CodedValue))
        self.assertTrue(isinstance(cv, float))

    def test_str_coded_value(self):
        cv = CodedValue('string', 'String')
        self.assertTrue(isinstance(cv, CodedValue))
        self.assertTrue(isinstance(cv, str))

    def test_equality(self):
        self.assertEqual(CodedValue(10, 'Integer'), 10)
        self.assertNotEqual(CodedValue(10, 'Integer'), 'Integer')


class TestDescription(unittest.TestCase):

    def test_init(self):
        with self.assertRaises(TypeError):
            D(25)

        with self.assertRaises(TypeError):
            D(None)

    def test_self_equality(self):
        self.assertEqual(D('Test'), D('Test'))
        self.assertNotEqual(D('Test 1'), D('Test 2'))

    def test_coded_value_equality(self):
        self.assertEqual(D('Test'), CodedValue(3, 'Test'))
        self.assertEqual(CodedValue(3, 'Test'), D('Test'))
        self.assertNotEqual(D('Test 1'), CodedValue(3, 'Test 2'))
        self.assertNotEqual(CodedValue(3, 'Test 2'), D('Test 1'))

    def test_coded_value_inequality(self):
        self.assertFalse(D('Test') != CodedValue(3, 'Test'))
        self.assertFalse(CodedValue(3, 'Test') != D('Test'))

    def test_other_equality(self):
        self.assertNotEqual(D('Test'), 'Test')


if __name__ == '__main__':
    unittest.main()
